{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "6eea777e",
   "metadata": {},
   "source": [
    "# ResNet (Residual Neural Networks)\n",
    "\n",
    "*Why did simply adding more layers to CNNs (like VGGNet or InceptionNet) fail to yield the expected performance gains—and sometimes even degraded accuracy?*\n",
    "\n",
    "Residual Neural Networks (ResNet) fundamentally changed the landscape of deep CNN training by introducing **residual connections** (a.k.a. skip connections). By stacking a series of **residual blocks**, ResNet enabled training CNNs with dozens or even hundreds of layers without succumbing to the **vanishing gradient problem**. Today, ResNet is considered one of the most important innovations in the history of deep learning, influencing architectures like **ResNeXt** and even **Transformers**.\n",
    "\n",
    "```{note}\n",
    "ResNeXt is an improvement over ResNet proposed by the same research group {footcite}`xie2017aggregated`. It widens the residual blocks via grouped convolutions, achieving higher performance without drastically increasing depth.\n",
    "```\n",
    "\n",
    "## Introduction and Context\n",
    "\n",
    "ResNet was introduced in {footcite}`he2016deep` to address a key challenge at the time: **CNNs deeper than about 20 layers were difficult to optimize and often performed worse than shallower counterparts**. Despite the success of VGGNet (16 or 19 layers) and InceptionNet, researchers still faced two major issues when pushing CNNs to 50 layers or more:\n",
    "\n",
    "1. **Degradation Problem**: Simply stacking more layers often *degraded* accuracy, rather than improving it.\n",
    "2. **Long Training Times**: Extremely deep CNNs took a long time to converge, especially if the network was prone to vanishing or exploding gradients.\n",
    "\n",
    "The ResNet solution was surprisingly simple yet groundbreaking: add skip connections that carry the original inputs across a few layers unmodified, letting the network focus on modeling the **residual**.\n",
    "\n",
    "## ResNet in Detail\n",
    "\n",
    "### Why Going Deeper Was Difficult\n",
    "\n",
    "*Shouldn't deeper networks always perform better because they have more parameters and expressive power?*\n",
    "\n",
    "In theory, **deeper** CNNs can capture richer, more complex patterns. However, two issues hindered progress:\n",
    "\n",
    "1. **Degradation Problem**\n",
    "   Even with techniques like batch normalization, adding more layers beyond ~20 caused training error to *increase*, not decrease. This phenomenon was *not* simply due to overfitting—rather, the deeper network failed to optimize properly.\n",
    "\n",
    "2. **Longer Training and Vanishing Gradients**\n",
    "   As more layers are added, gradients can vanish (or explode). Backprop had trouble sending meaningful error signals all the way to early layers, causing them to learn slowly or not at all.\n",
    "\n",
    "### Key Proposal: Residual Learning with Skip Connections\n",
    "\n",
    "*What if each stack of layers simply learned a correction (residual) to the identity mapping?*\n",
    "\n",
    "A **residual block** consists of two (or three) convolutions grouped together, plus a **skip connection**:\n",
    "\n",
    "- **Residual Path**: A few convolution layers (for example, two 3×3 conv layers) modeling a function $ F(\\mathbf{x}) $.\n",
    "- **Skip (Identity) Path**: A direct path for $\\mathbf{x}$ to bypass the convolutions entirely.\n",
    "\n",
    "At the end of the block, the skip path is added elementwise to the residual path:\n",
    "$$\n",
    "\\mathbf{y} = F(\\mathbf{x}) + \\mathbf{x}.\n",
    "$$\n",
    "\n",
    "In PyTorch, you can implement a basic residual block as follows:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "5aa17efe",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "import torch.nn as nn\n",
    "\n",
    "class BasicBlock(nn.Module):\n",
    "    def __init__(self, in_channels, out_channels, stride=1):\n",
    "        super().__init__()\n",
    "        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)\n",
    "        self.bn1 = nn.BatchNorm2d(out_channels)\n",
    "        self.relu = nn.ReLU(inplace=True)\n",
    "\n",
    "        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)\n",
    "        self.bn2 = nn.BatchNorm2d(out_channels)\n",
    "\n",
    "    def forward(self, x):\n",
    "        identity = x\n",
    "\n",
    "        out = self.conv1(x)\n",
    "        out = self.bn1(out)\n",
    "        out = self.relu(out)\n",
    "\n",
    "        out = self.conv2(out)\n",
    "        out = self.bn2(out)\n",
    "\n",
    "        out += identity\n",
    "        out = self.relu(out)\n",
    "        return out"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b41cf4cb",
   "metadata": {},
   "source": [
    "```{figure} https://www.researchgate.net/publication/364330795/figure/fig7/AS:11431281176036099@1689999593116/Basic-residual-block-of-ResNet.png\n",
    ":name: resnet-block\n",
    ":align: center\n",
    ":width: 50%\n",
    "A basic 2-layer residual block (left) vs. a plain block without skip (right). The skip connection allows the input $\\mathbf{x}$ to directly add to the block’s output.\n",
    "```\n",
    "\n",
    "By stacking many such blocks, the network effectively *cascades* small residual changes across layers. The key benefits are:\n",
    "\n",
    "1. **Easier Optimization**\n",
    "   Instead of learning a full mapping $\\mathbf{y} = G(\\mathbf{x})$, the block learns only the difference $G(\\mathbf{x}) - \\mathbf{x}$. This decomposition often proves easier to optimize.\n",
    "\n",
    "   ```{note}\n",
    "   If the optimal mapping is close to identity (i.e., the layer isn't very important), the network can easily \"skip\" it by learning $F(\\mathbf{x}) \\approx 0$. If a more complex transformation is needed, the residual path can still learn it. This makes training more robust—the network doesn’t have to work as hard to preserve important information through deep layers.\n",
    "   ```\n",
    "\n",
    "2. **Ensemble-Like Behavior**\n",
    "   When you chain $N$ residual blocks, you effectively create numerous paths for gradient flow—some skip many layers, some pass through multiple convolutions. This variety of gradient routes can speed convergence and reduce the risk of vanishing gradients {footcite}`veit2016residual`.\n",
    "\n",
    "    ```{figure} https://arxiv.org/html/2405.01725v1/x28.png\n",
    "    :name: resnet-gradient-flow\n",
    "    :align: center\n",
    "    :width: 100%\n",
    "    The gradient flow in ResNet with skip connections.\n",
    "    ```\n",
    "\n",
    "3. **Deeper Without Degradation**\n",
    "   ResNet-50, -101, and -152 can be trained without suffering the performance drop typical of overly deep “plain” networks.\n",
    "\n",
    "### Bottleneck Blocks for Deep ResNet\n",
    "\n",
    "ResNet has some variants depending on the depth. For deep ResNet, the **bottleneck** design is used to maintain computational efficiency.\n",
    "\n",
    "```{figure} https://i.sstatic.net/kbiIG.png\n",
    ":name: resnet-bottleneck-block\n",
    ":align: center\n",
    ":width: 80%\n",
    "A bottleneck block of ResNet.\n",
    "```\n",
    "\n",
    "This bottleneck block consists of three convolutions instead of two, where:\n",
    "- the first $1 \\times 1$ conv **reduces** the feature dimension.\n",
    "- the second $3 \\times 3$ conv operates on this reduced dimension.\n",
    "- the third $1 \\times 1$ conv **restores** the dimension.\n",
    "\n",
    "This approach shrinks the intermediate feature map, saving computational cost while retaining overall representational capacity. It was inspired by InceptionNet’s “bottleneck” idea {footcite}`szegedy2016inception,szegedy2015going`.\n",
    "\n",
    "```{tip}\n",
    "**ResNet-50**, **ResNet-101**, and **ResNet-152** all use bottleneck blocks. While they have more layers, they remain computationally feasible and yield progressively better accuracy on ImageNet.\n",
    "```\n",
    "\n",
    "## ResNeXt: A ResNet Improvement\n",
    "\n",
    "*What if we can widen the residual blocks without drastically increasing overall parameters?*\n",
    "\n",
    "**ResNeXt** {footcite}`xie2017aggregated` is an evolution of ResNet that:\n",
    "1. **Splits** the bottleneck conv pathway into multiple “cardinality” groups (e.g., 32 groups).\n",
    "2. **Aggregates** those parallel paths (grouped convolutions) back into a single output.\n",
    "\n",
    "By increasing **cardinality** (the number of parallel conv groups) instead of just adding more channels or layers, ResNeXt achieves better accuracy with moderate complexity. This approach also draws on the idea of Inception’s multi-branch parallel conv, but unifies them into a single grouped-convolution block.\n",
    "\n",
    "```{figure} https://production-media.paperswithcode.com/methods/Screen_Shot_2020-06-06_at_4.32.52_PM.png\n",
    ":name: resnext-block\n",
    ":align: center\n",
    ":width: 50%\n",
    "A basic block of ResNeXt, showing multiple grouped-conv “paths” that are aggregated.\n",
    "```\n",
    "\n",
    "## Implementation of ResNet\n",
    "\n",
    "- [Writing ResNet from Scratch in PyTorch](https://www.digitalocean.com/community/tutorials/writing-resnet-from-scratch-in-pytorch)\n",
    "\n",
    "## Summary\n",
    "\n",
    "1. **Residual Learning**\n",
    "   ResNet overcame the degradation problem by framing deeper CNNs as a series of residual blocks, each learning a function $ F(\\mathbf{x}) $ that is added to $\\mathbf{x}$.\n",
    "\n",
    "2. **Scalability**\n",
    "   With skip connections, ResNet-50, -101, and -152 exhibit higher accuracy without the optimization collapse typical of deeper plain networks.\n",
    "\n",
    "3. **Bottleneck & Beyond**\n",
    "   For high-depth architectures, the bottleneck design $(1\\times1 \\to 3\\times3 \\to 1\\times1)$ improves efficiency. ResNeXt further extends ResNet by widening these pathways via grouped convolutions.\n",
    "\n",
    "4. **Lasting Impact**\n",
    "   Residual connections are now ubiquitous—not just in CNNs but also in Transformers, large-scale language models, U-Nets, and many other architectures. They simplify optimization and significantly improve gradient flow in very deep models.\n",
    "\n",
    "```{note}\n",
    "ResNet’s simplicity made it a foundation for many follow-up architectures. Unlike designs with complex branching (e.g., Inception blocks), ResNet remains easy to implement, debug, and extend—an important factor behind its widespread adoption.\n",
    "```\n",
    "\n",
    "## Suggested Exercises\n",
    "\n",
    "1. **Implement a Basic (Non-Bottleneck) Residual Block**\n",
    "   - Create a two-convolution block with skip connections.\n",
    "   - Test it on random data to confirm dimensions match.\n",
    "2. **Train a Small ResNet**\n",
    "   - Implement ResNet-18 or ResNet-34 from scratch on a smaller dataset (e.g., CIFAR-10).\n",
    "   - Observe the training curve and compare to a plain CNN of the same depth.\n",
    "3. **Experiment with Bottleneck Blocks**\n",
    "   - Convert your ResNet-34 to a bottleneck-based ResNet-50-like structure.\n",
    "   - Check the parameter count and performance difference on CIFAR-10 or a subset of ImageNet.\n",
    "\n",
    "```{footbibliography} references.bib\n",
    ":style: unsrt\n",
    ":filter: docname in docnames\n",
    "```"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "formats": "md:myst",
   "text_representation": {
    "extension": ".md",
    "format_name": "myst"
   }
  },
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.9"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}